#include "imageloader.h"
#include <qimage.h>
#include <stdlib.h>

#define CONVERT_SHORT(l, m) (((m)<<8)|l)
#define CONVERT_INT(a, b, c, d) (((d) << 24)|((c) << 16)|((b) << 8) | a)

const int BMP_FILEHDR_SIZE = 14;		// size of BMP_FILEHDR data
const int BMP_OLD  = 12;			// old Windows/OS2 BMP size
const int BMP_WIN  = 40;			// new Windows BMP size
const int BMP_OS2  = 64;			// new OS/2 BMP size
const int BMP_RGB  = 0;				// no compression
const int BMP_RLE8 = 1;				// run-length encoded, 8 bits
const int BMP_RLE4 = 2;				// run-length encoded, 4 bits
const int BMP_BITFIELDS = 3;			// RGB values encoded in data as bit-fields

class BMP_FILEHDR {				// BMP file header
public:
    BMP_FILEHDR(){;}
    char   bfType[2];				// "BM"
    Q_INT32  bfSize;				// size of file
    Q_INT16  bfReserved1;
    Q_INT16  bfReserved2;
    Q_INT32  bfOffBits;				// pointer to the pixmap bits
};


class BMP_INFOHDR {
public:
    BMP_INFOHDR(){;}
    // BMP information header
    Q_INT32  biSize;				// size of this struct
    Q_INT32  biWidth;				// pixmap width
    Q_INT32  biHeight;				// pixmap height
    Q_INT16  biPlanes;				// should be 1
    Q_INT16  biBitCount;			// number of bits per pixel
    Q_INT32  biCompression;			// compression method
    Q_INT32  biSizeImage;				// size of image
    Q_INT32  biXPelsPerMeter;			// horizontal resolution
    Q_INT32  biYPelsPerMeter;			// vertical resolution
    Q_INT32  biClrUsed;				// number of colors used
    Q_INT32  biClrImportant;			// number of important colors
};

BMPImageLoader::BMPImageLoader(QImage &image, ImageView *view)
    : ImageLoader(image, view)
{
    state = FileHeader;
    bf = new BMP_FILEHDR();
    bi = new BMP_INFOHDR();
    count = 0;
    position = 0;
    temp = 0;
    hasError = false;

    red_shift = 0;
    green_shift = 0;
    blue_shift = 0;
    red_scale = 0;
    green_scale = 0;
    blue_scale = 0;
    ncols = 0;
    depth = 0;
    currentX = 0;
    currentY = 0;
    inEsc = false;
    img = &image;
    hold = (unsigned char *)malloc(40);
}

BMPImageLoader::~BMPImageLoader()
{
    delete bf;
    delete bi;
    free(hold);
}

void BMPImageLoader::createImage()
{
    switch(nbits){
	case 32:
	case 24:
	case 16:
	    depth = 32;
	    break;
	case 8:
	case 4:
	    depth = 8;
	    break;
	default:
            depth = 1;
    }
    if(depth == 32)				// there's no colormap
        ncols = 0;
    else					// # colors used
        ncols = bi->biClrUsed ? bi->biClrUsed : 1 << nbits;

    img->create(w, h, depth, ncols, nbits == 1 ?
                QImage::BigEndian : QImage::IgnoreEndian);
    img->setDotsPerMeterX(bi->biXPelsPerMeter);
    img->setDotsPerMeterY(bi->biYPelsPerMeter);
    data = (unsigned int *)img->scanLine(h-1);
    byteData = (unsigned char *)img->scanLine(h-1);
    currentY = h-1;
    bpl = img->bytesPerLine();
    if(nbits == 16 || nbits == 24 || nbits == 32){
        bpl = ((w*nbits+31)/32)*4;
        free(hold);
        hold = (unsigned char *)malloc(bpl);
        count = 0;
    }
    qWarning("Created %dx%d, %dbpp image, bpl: %d", w, h, depth,
             img->bytesPerLine());
}

void BMPImageLoader::calcImageType()
{
    if(nbits == 1){
        state = ImageBitmap;
        qWarning("Image Type: bitmap");
    }
    else if(nbits == 4){
        if(comp == BMP_RLE4)
            state = Image4BitRLE;
        else
            state = Image4Bit;
        qWarning("Image Type: %s", comp == BMP_RLE4 ? "4bit RLE" : "4bit");
    }
    else if(nbits == 8){
        if(comp == BMP_RLE8)
            state = Image8BitRLE;
        else
            state = Image8Bit;
        qWarning("Image Type: %s", comp == BMP_RLE8 ? "8bit RLE" : "8bit");
    }
    else if(nbits == 16){
        state = Image16Bit;
        qWarning("Image Type: 16bit");
    }
    else if(nbits == 24){
        state = Image24Bit;
        qWarning("Image Type: 24bit");
    }
    else if(nbits == 32){
        state = Image32Bit;
        qWarning("Image Type: 32bit");
    }

    else{
        qWarning("Whoa, unknown image type!");
        state = Error;
    }
}

bool BMPImageLoader::readData(char *buffer, int length)
{
    int lastUpdatedRow=-1;

    while(length){
        --length;
        unsigned char ch = *buffer++;

        switch(state){
        case FileHeader:
            hold[count++] = ch;
            if(count == BMP_FILEHDR_SIZE + 4){
                bf->bfType[0] = hold[0];
                bf->bfType[1] = hold[1];
                bf->bfSize = CONVERT_INT(hold[2], hold[3], hold[4], hold[5]);
                bf->bfReserved1 = CONVERT_SHORT(hold[6], hold[7]);
                bf->bfReserved2 = CONVERT_SHORT(hold[8], hold[9]);
                bf->bfOffBits = CONVERT_INT(hold[10], hold[11],
                                            hold[12], hold[13]);
                // preload info header type
                bi->biSize = CONVERT_INT(hold[14], hold[15],
                                         hold[16], hold[17]);
                if(bi->biSize == BMP_WIN || bi->biSize == BMP_OS2)
                    state = InfoHeader;
                else
                    state = InfoHeaderOld;
                count = 0;

                if(qstrncmp(bf->bfType, "BM", 2) != 0){
                    qWarning("Improper BMP ID!");
                    state = Error;
                    return(false);
                }
                else
                    qWarning("Proper BMP ID");
            }
            break;
        case InfoHeader:
            hold[count++] = ch;
            if(count == 36){
                qWarning("Got info header!");
                bi->biWidth = CONVERT_INT(hold[0], hold[1],
                                          hold[2], hold[3]);
                bi->biHeight = CONVERT_INT(hold[4], hold[5],
                                           hold[6], hold[7]);
                bi->biPlanes = CONVERT_SHORT(hold[8], hold[9]);
                bi->biBitCount = CONVERT_SHORT(hold[10], hold[11]);
                bi->biCompression = CONVERT_INT(hold[12], hold[13],
                                                hold[14], hold[15]);
                bi->biSizeImage = CONVERT_INT(hold[16], hold[17],
                                              hold[18], hold[19]);
                bi->biXPelsPerMeter = CONVERT_INT(hold[20], hold[21],
                                                  hold[22], hold[23]);
                bi->biYPelsPerMeter = CONVERT_INT(hold[24], hold[25],
                                                  hold[26], hold[27]);
                bi->biClrUsed = CONVERT_INT(hold[28], hold[29],
                                            hold[30], hold[31]);
                bi->biClrImportant = CONVERT_INT(hold[32], hold[33],
                                                 hold[34], hold[35]);

                w = bi->biWidth;
                h = bi->biHeight;
                nbits = bi->biBitCount;
                t = bi->biSize;
                comp = bi->biCompression;
                state = SkipToColorMap;
                count = 0;
                if(!(nbits == 1 || nbits == 4 || nbits == 8 || nbits == 16 || nbits == 24 || nbits == 32) ||
                   bi->biPlanes != 1 || comp > BMP_BITFIELDS){
                    qWarning("Unknown BMP image!");
                    hasError = true;
                    state = Error;
                    return(false);
                }
                if(!(comp == BMP_RGB || (nbits == 4 && comp == BMP_RLE4) ||
                     (nbits == 8 && comp == BMP_RLE8) || ((nbits == 16 || nbits == 32) && comp == BMP_BITFIELDS))){
                    qWarning("Unknown BMP compression!");
                    hasError = true;
                    state = Error;
                    return(false);
                }
                if(comp == BMP_RLE8 || comp == BMP_RLE4){
                    // no support for incremental RLE yet because it can
                    // jump around in the file
                    qWarning("No support for incremental RLE");
                    hasError = true;
                    state = Error;
                    return(false);
                }
                createImage();
            }
            break;
        case InfoHeaderOld:
            qWarning("Old header!");
            hold[count++] = ch;
            if(count == 8){
                Q_INT16 tmpW, tmpH;
                tmpW = CONVERT_SHORT(hold[0], hold[1]);
                tmpH = CONVERT_SHORT(hold[2], hold[3]);
                bi->biPlanes = CONVERT_SHORT(hold[4], hold[5]);
                bi->biBitCount = CONVERT_SHORT(hold[6], hold[7]);
                bi->biWidth = tmpW;
                bi->biHeight = tmpH;
                bi->biCompression = BMP_RGB;
                bi->biSizeImage = 0;
                bi->biXPelsPerMeter = bi->biYPelsPerMeter = 0;
                bi->biClrUsed = bi->biClrImportant = 0;

                w = bi->biWidth;
                h = bi->biHeight;
                nbits = bi->biBitCount;
                t = bi->biSize;
                comp = bi->biCompression;
                state = SkipToColorMap;
                count = 0;

                if(!(nbits == 1 || nbits == 4 || nbits == 8 || nbits == 16 || nbits == 24 || nbits == 32) ||
                   bi->biPlanes != 1 || comp > BMP_BITFIELDS){
                    qWarning("Unknown BMP image!");
                    hasError = true;
                    state = Error;
                    return(false);
                }
                if(!(comp == BMP_RGB || (nbits == 4 && comp == BMP_RLE4) ||
                     (nbits == 8 && comp == BMP_RLE8) || ((nbits == 16 || nbits == 32) && comp == BMP_BITFIELDS))){
                    qWarning("Unknown BMP compression!");
                    hasError = true;
                    state = Error;
                    return(false);
                }
                createImage();
            }
            break;
        case SkipToColorMap:
            if(position == BMP_FILEHDR_SIZE + bi->biSize){
                qWarning("At colormap, skipped %d bytes", temp);
                temp = 0;
                if(ncols){
                    colormapSize = ncols * (t == BMP_OLD ? 3 : 4);
                    if(colormapSize > 40){
                        free(hold);
                        hold = (unsigned char *)malloc(colormapSize);
                    }
                    --buffer;
                    --position;
                    ++length;
                    state = ColorMap;
                    qWarning("Colormapped image");
                }
                else if(comp == BMP_BITFIELDS && (nbits == 16 || nbits == 32)){
                    --buffer;
                    --position;
                    ++length;
                    state = BitFields;
                    qWarning("Bitfield image");
                }
                else if(comp == BMP_RGB && (nbits == 24 || nbits == 32)){
                    blue_mask = 0x000000ff;
                    green_mask = 0x0000ff00;
                    red_mask = 0x00ff0000;
                    blue_shift = 0;
                    green_shift = 8;
                    red_shift = 16;
                    blue_scale = green_scale = red_scale = 1;
                    state = SkipToImage;
                }
                else if(comp == BMP_RGB && nbits == 16){
                    qWarning("RGB values for 15/16 bpp not supported!");
                    hasError = true;
                    state = Error;
                    return(false);
                }
                else{
                    qWarning("Whoa, don't know the color map type!");
                    hasError = true;
                    state = Error;
                    return(false);
                }
            }
            else
                ++temp;
            break;
        case SkipToImage:
            if(bf->bfOffBits > 0 && bf->bfOffBits >= position){
                if(position == bf->bfOffBits){
                    qWarning("At image, skipped %d bytes", temp);
                    --buffer;
                    --position;
                    ++length;
                    calcImageType();
                }
                else
                    ++temp;
            }
            else{
                qWarning("No need to skip bytes to get to image, pos: %d",
                         position);
                --buffer;
                --position;
                ++length;
                calcImageType();
            }
            break;
        case ColorMap:
            hold[count++] = ch;
            if(count == colormapSize-1){
                if(t == BMP_OLD){
                    qWarning("Using old colormap type");
                    for(int i=0, c=0; i < colormapSize-1; i+=3, ++c){
                        img->setColor(c, qRgb(hold[i+2], hold[i+1], hold[i]));
                    }
                }
                else{
                    qWarning("Using new colormap type");
                    int i, c;
                    for(i=0, c=0; i < colormapSize-1; i+=4, ++c){
                        img->setColor(c, qRgb(hold[i+2], hold[i+1], hold[i]));
                    }
                    qWarning("Loaded %d colors", c);
                }
                state = SkipToImage;
            }
            break;
        case BitFields:
            hold[count++] = ch;
            if(count == 12){
                red_mask = CONVERT_INT(hold[0], hold[1], hold[2], hold[3]);
                green_mask = CONVERT_INT(hold[4], hold[5], hold[6], hold[7]);
                blue_mask = CONVERT_INT(hold[8], hold[9], hold[10], hold[11]);
                red_shift = calc_shift(red_mask);
                red_scale = 256/((red_mask >> red_shift) + 1);
                green_shift = calc_shift(green_mask);
                green_scale = 256/((green_mask >> green_shift) + 1);
                blue_shift = calc_shift(blue_mask);
                blue_scale = 256/((blue_mask >> blue_shift) + 1);
                count = 0;
                state = SkipToImage;
            }
            break;
        case Image16Bit:
        case Image24Bit:
        case Image32Bit:
            hold[count++] = ch;
            if(count == bpl){
                int c;
                unsigned int *p = (unsigned int *)img->scanLine(currentY);
                for(int i=0; i < bpl; i += nbits/8){
                    c = hold[i+2] | hold[i] <<8;
                    if(nbits != 16)
                        c |=  hold[i+1] <<16;
                    *p++ = qRgb(((c & red_mask) >> red_shift) * red_scale,
                                ((c & green_mask) >> green_shift) * green_scale,
                                ((c & blue_mask) >> blue_shift) * blue_scale);
                }
                lastUpdatedRow = currentY;
                --currentY;
                if(currentY  < 0)
                    state = End;
                count = 0;
            }
            break;
        case Image4Bit:
            byteData[currentX++] = ch >> 4;
            byteData[currentX++] = ch & 0x0F;
            if(currentX >= bpl){
                currentX = 0;
                lastUpdatedRow = currentY;
                --currentY;
                if(currentY < 0){
                    state = End;
                }
                else{
                    byteData = (unsigned char *)img->scanLine(currentY);
                }
            }
            break;
        case Image8Bit:
        case ImageBitmap:{
            byteData[currentX] = ch;
            ++currentX;
            if(currentX == bpl){
                currentX = 0;
                lastUpdatedRow = currentY;
                --currentY;
                if(currentY < 0){
                    state = End;
                }
                else{
                    byteData = (unsigned char *)img->scanLine(currentY);
                }
            }
            break;
        }
        case Error:
            hasError = true;
            return(false);
            break;
        case End:
            return(true);
            break;
        default:
            break;
        }
        ++position;
    }
    return(true);
}

bool BMPImageLoader::error()
{
    return(hasError);
}

int BMPImageLoader::calc_shift(int mask)
{
    int result = 0;
    while (!(mask & 1)) {
	result++;
	mask >>= 1;
    }
    return result;
}

void BMPImageLoader::swapColormap()
{
    int i;
    if(img->depth() == 1 && img->numColors() == 2){
        register uint *p = (uint *)img->bits();
        int nbytes = img->numBytes();
        for(i=0; i<nbytes/4; i++){
            *p = ~*p;
            p++;
        }
        uchar *p2 = (uchar *)p;
        for ( i=0; i<(nbytes&3); i++ ) {
            *p2 = ~*p2;
	    p2++;
        }
	QRgb t = img->color(0);		// swap color 0 and 1
	img->setColor( 0, img->color(1) );
	img->setColor( 1, t );
    }
}


