#include "imageloader.h"
#include <stdio.h>
#include <setjmp.h>
#include <qdatetime.h>

JPEGImageLoader *currentJPEGLoader;
QImage *currentJPEGImage;
static const int max_buf = 8192;
static const int max_consumingtime = 500;
bool hasJPEGError, hasJPEGHeader;

extern "C" {
#define XMD_H
#include <jpeglib.h>
#undef const

    struct mosfet_error_mgr : public jpeg_error_mgr{
        jmp_buf setjmp_buffer;
    };

    static void mosfet_error_exit (j_common_ptr cinfo){
        mosfet_error_mgr* myerr = (mosfet_error_mgr*) cinfo->err;
        char buffer[JMSG_LENGTH_MAX];
        (*cinfo->err->format_message)(cinfo, buffer);
        qWarning(buffer);
        longjmp(myerr->setjmp_buffer, 1);
    }

    struct mosfet_jpeg_source_mgr : public jpeg_source_mgr {
        JOCTET buffer[max_buf];

        int valid_buffer_len;
        unsigned int skip_input_bytes;
        int ateof;
        QTime decoder_timestamp;
        bool final_pass;
        bool decoding_done;

    public:
        mosfet_jpeg_source_mgr();
    };

    static void mosfet_j_decompress_dummy(j_decompress_ptr){
        ;
    }

    static boolean mosfet_fill_input_buffer(j_decompress_ptr cinfo){
        mosfet_jpeg_source_mgr* src = (mosfet_jpeg_source_mgr*)cinfo->src;
        if(src->ateof){
            /* Insert a fake EOI marker - as per jpeglib recommendation */
            src->buffer[0] = (JOCTET) 0xFF;
            src->buffer[1] = (JOCTET) JPEG_EOI;
            src->bytes_in_buffer = 2;
            return(true);
        }
        else
            return(false);  /* I/O suspension mode */
    }

    static void mosfet_skip_input_data(j_decompress_ptr cinfo, long num_bytes){
        if(num_bytes <= 0)
            return; /* required noop */
        mosfet_jpeg_source_mgr* src = (mosfet_jpeg_source_mgr*)cinfo->src;
        src->skip_input_bytes += num_bytes;

        unsigned int skipbytes = QMIN(src->bytes_in_buffer, src->skip_input_bytes);
        if(skipbytes < src->bytes_in_buffer)
            memcpy(src->buffer, src->next_input_byte+skipbytes, src->bytes_in_buffer - skipbytes);

        src->bytes_in_buffer -= skipbytes;
        src->valid_buffer_len = src->bytes_in_buffer;
        src->skip_input_bytes -= skipbytes;

        /* adjust data for jpeglib */
        cinfo->src->next_input_byte = (JOCTET *) src->buffer;
        cinfo->src->bytes_in_buffer = (size_t) src->valid_buffer_len;
    }

    mosfet_jpeg_source_mgr::mosfet_jpeg_source_mgr(){
        jpeg_source_mgr::init_source = mosfet_j_decompress_dummy;
        jpeg_source_mgr::fill_input_buffer = mosfet_fill_input_buffer;
        jpeg_source_mgr::skip_input_data = mosfet_skip_input_data;
        jpeg_source_mgr::resync_to_restart = jpeg_resync_to_restart;
        jpeg_source_mgr::term_source = mosfet_j_decompress_dummy;
        bytes_in_buffer = 0;
        valid_buffer_len = 0;
        skip_input_bytes = 0;
        ateof = 0;
        next_input_byte = buffer;
        final_pass = false;
        decoding_done = false;
    }
}

class JPEGPrivate
{
public:
    JPEGPrivate(){;}

    struct jpeg_decompress_struct cinfo;
    struct mosfet_error_mgr jerr;
    struct mosfet_jpeg_source_mgr jsrc;
};

JPEGImageLoader::JPEGImageLoader(QImage &image, ImageView *view)
    : ImageLoader(image, view)
{
    d = new JPEGPrivate();
    currentJPEGImage = &image;
    currentJPEGLoader = this;
    hasJPEGError = false;
    hasJPEGHeader = false;

    memset(&d->cinfo, 0, sizeof(d->cinfo));
    d->cinfo.err = jpeg_std_error(&d->jerr);
    jpeg_create_decompress(&d->cinfo);
    d->cinfo.err = jpeg_std_error(&d->jerr);
    d->jerr.error_exit = mosfet_error_exit;
    d->cinfo.src = &d->jsrc;
    state = Init;
}

JPEGImageLoader::~JPEGImageLoader()
{
    jpeg_destroy_decompress(&d->cinfo);
    delete d;
}

bool JPEGImageLoader::readData(char *buffer, int size)
{

    if(hasJPEGError){
        qWarning("Has Error!");
        return(false);
    }
    int consumed = processData(buffer, size);
    while(consumed && size > 0){
        buffer = buffer+consumed;
        size -= consumed;
        if(size > 0)
           consumed = processData(buffer, size);
    }
    return(true);

}

int JPEGImageLoader::processData(char *buffer, int length)
{
    if(d->jsrc.ateof)
        return(true);
    if(setjmp(d->jerr.setjmp_buffer)){
        hasJPEGError = true;
        return(false);
    }

    int consumed = QMIN(length, max_buf - d->jsrc.valid_buffer_len);
    memcpy(d->jsrc.buffer + d->jsrc.valid_buffer_len, buffer, consumed);
    d->jsrc.valid_buffer_len += consumed;

    if(d->jsrc.skip_input_bytes){
        int skipbytes = QMIN((unsigned) d->jsrc.valid_buffer_len, d->jsrc.skip_input_bytes);
        if(skipbytes < d->jsrc.valid_buffer_len)
            memcpy(d->jsrc.buffer, d->jsrc.buffer+skipbytes, d->jsrc.valid_buffer_len - skipbytes);

        d->jsrc.valid_buffer_len -= skipbytes;
        d->jsrc.skip_input_bytes -= skipbytes;
        // still more bytes to skip
        if(d->jsrc.skip_input_bytes){
            if(consumed <= 0)
                qWarning("Error in JPEGImageLoader::readData");
            return(consumed);
        }
    }

    d->cinfo.src->next_input_byte = (JOCTET *) d->jsrc.buffer;
    d->cinfo.src->bytes_in_buffer = (size_t) d->jsrc.valid_buffer_len;

    if(state == Init){
        if(jpeg_read_header(&d->cinfo, true) != JPEG_SUSPENDED){
            hasJPEGHeader = true;
            state = startDecompress;
        }
    }
    if(state == startDecompress){
        d->cinfo.buffered_image = TRUE;
        d->cinfo.do_fancy_upsampling = FALSE;
        d->cinfo.do_block_smoothing = FALSE;
        d->cinfo.dct_method = JDCT_FASTEST;
        // FALSE: IO suspension
        if(jpeg_start_decompress(&d->cinfo)){
            if(d->cinfo.output_components == 3 || d->cinfo.output_components == 4){
                currentJPEGImage->create(d->cinfo.output_width, d->cinfo.output_height, 32);
                currentJPEGImage->fill(qRgb(255, 255, 255));
            }
            else if( d->cinfo.output_components == 1){
                currentJPEGImage->create(d->cinfo.output_width, d->cinfo.output_height, 8, 256);
                for(int i=0; i<256; i++)
                    currentJPEGImage->setColor(i, qRgb(i, i, i));
                currentJPEGImage->fill(255);
            }
            d->jsrc.decoder_timestamp.start();
            state = decompressStarted;
        }
    }
    if(state == decompressStarted){
        state = (!d->jsrc.final_pass && d->jsrc.decoder_timestamp.elapsed() < max_consumingtime)
            ? consumeInput : prepareOutputScan;
    }
    if(state == consumeInput){
        int retval;
        do{
            retval = jpeg_consume_input(&d->cinfo);
        } while(retval != JPEG_SUSPENDED && retval != JPEG_REACHED_EOI);
        if(d->jsrc.decoder_timestamp.elapsed() > max_consumingtime ||
           d->jsrc.final_pass || retval == JPEG_REACHED_EOI ||
           retval == JPEG_REACHED_SOS)
            state = prepareOutputScan;
        else
            state = consumeInput;
    }
    if(state == prepareOutputScan){
        d->jsrc.decoder_timestamp.restart();
        d->cinfo.buffered_image = TRUE;
        jpeg_start_output(&d->cinfo, d->cinfo.input_scan_number);
        state = doOutputScan;
    }
    if(state == doOutputScan){
        if(currentJPEGImage->isNull() || d->jsrc.decoding_done){
            return(consumed);
        }
        uchar** lines = currentJPEGImage->jumpTable();
        int oldoutput_scanline = d->cinfo.output_scanline;
        while(d->cinfo.output_scanline < d->cinfo.output_height &&
              jpeg_read_scanlines(&d->cinfo, lines+d->cinfo.output_scanline, d->cinfo.output_height))
            ; // here happens all the magic of decoding

        int completed_scanlines = d->cinfo.output_scanline - oldoutput_scanline;
        if(d->cinfo.output_components == 3){
            // Expand 24->32 bpp.
            for(int j=oldoutput_scanline; j<oldoutput_scanline+completed_scanlines; j++) {
                uchar *in = currentJPEGImage->scanLine(j) + d->cinfo.output_width * 3;
                QRgb *out = (QRgb*)currentJPEGImage->scanLine(j);
                for (uint i=d->cinfo.output_width; i--;){
                    in-=3;
                    out[i] = qRgb(in[0], in[1], in[2]);
                }
            }
        }
        if(completed_scanlines){
            //QRect r(0, oldoutput_scanline, d->cinfo.output_width, completed_scanlines);
            currentJPEGLoader->imageUpdated(oldoutput_scanline,
                                            oldoutput_scanline+completed_scanlines);

        }

        if(d->cinfo.output_scanline >= d->cinfo.output_height){
            jpeg_finish_output(&d->cinfo);
            d->jsrc.final_pass = jpeg_input_complete(&d->cinfo);
            d->jsrc.decoding_done = d->jsrc.final_pass && d->cinfo.input_scan_number == d->cinfo.output_scan_number;
            if(!d->jsrc.decoding_done){
                // don't return until necessary!
                d->jsrc.decoder_timestamp.restart();
                state = decompressStarted;
            }
        }
        if(state == doOutputScan && d->jsrc.decoding_done) {
            (void) jpeg_finish_decompress(&d->cinfo);
            (void) jpeg_destroy_decompress(&d->cinfo);
            // image is completed here
            d->jsrc.ateof = true;
            currentJPEGLoader->imageUpdated(0, d->cinfo.output_height);
            return(0);
        }
    }
    if(d->jsrc.bytes_in_buffer)
        memcpy(d->jsrc.buffer, d->jsrc.next_input_byte, d->jsrc.bytes_in_buffer);
    d->jsrc.valid_buffer_len = d->jsrc.bytes_in_buffer;
    return(consumed);
}

bool JPEGImageLoader::error()
{
    return(hasJPEGError);
}


